"""Module to support lazy evaluation of values."""

import importlib
import logging

from collections.abc import Callable, Iterator, Mapping, MutableMapping
from types import ModuleType
from typing import Any, TypeVar


_logger = logging.getLogger(__name__)


K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")


class LazyMap(MutableMapping[K, V]):
    """
    A mapping of key-value pairs, in which a value can be lazily initialized at the time of
    first access. This is useful to allow resources to access other resources, preventing
    circular imports.

    Parameters:
    • init: mapping of key-value pairs to initialize the lazy map

    To lazily initialize a value, set it to a no-argument function callback that has been
    decorated with the @lazy decorator. When the value is then first accessed, the callback
    will be called to initialize the value. The resulting value will then be stored in the
    mapping.

    Because values in the map are lazily initialized, this map will not compare equal to any
    other object.
    """

    def __init__(self, init: Mapping[K, V] | None = None):
        super().__init__()
        self._store = {}
        if init is not None:
            for key, value in init.items():
                self[key] = value

    def __getitem__(self, key: K) -> V:
        value = self._store[key]
        while is_lazy(value):
            self._store[key] = value = value()
        return value

    def __setitem__(self, key: K, value: V) -> None:
        self._store[key] = value

    def __delitem__(self, key: K) -> None:
        del self._store[key]

    def __iter__(self) -> Iterator[K]:
        return iter(self._store)

    def __len__(self) -> int:
        return len(self._store)

    def __contains__(self, item: V) -> bool:
        return item in self._store


class LazySimpleNamespace:
    """
    An object that provides attribute access to its namespace, whose values are lazily
    initialized at time of access.

    Parameters:
    • kwargs: key-value pairs to initialize the lazy map

    To lazily initialize a value, set it to a no-argument function callback that has been
    decorated with the @lazy decorator. When the value is then first accessed, the callback
    will be called to initialize the value. The resulting value will then be stored in the
    namespace.

    Because values in this namespace are lazily initialized, this namespace will not compare
    equal to any other object.
    """

    def __init__(self, **kwargs):
        self._fondat__lazymap = LazyMap(kwargs)

    def __getattr__(self, name):
        try:
            return self._fondat__lazymap[name]
        except KeyError:
            raise AttributeError

    def __setattr__(self, name, value):
        if name == "_fondat__lazymap" or name in object.__dir__(self):
            object.__setattr__(self, name, value)
        else:
            self._fondat__lazymap[name] = value

    def __delattr__(self, name):
        if name in object.__dir__(self):
            object.__delattr__(self, name)
        else:
            try:
                del self._fondat__lazymap[name]
            except KeyError:
                raise AttributeError

    def __dir__(self):
        return [*object.__dir__(self), *self._fondat__lazymap.keys()]


def lazy(function: T) -> T:
    """Decorate a function to tag it as a lazy initializer."""
    if not callable(function):
        raise TypeError("function must be callable")
    setattr(function, "_fondat__lazy", True)
    return function


def lazy_import(module_name: str) -> Callable[[], ModuleType]:
    """
    Return a lazy callback function that imports a module and returns it.

    Parameters:
    • module_name: the name of the module to import
    """

    def callback():
        return importlib.import_module(module_name)

    return lazy(callback)


def lazy_import_attr(module_name: str, attr_name: str) -> Callable[[], Any]:
    """
    Return a lazy callback function that imports a module and returns an attribute from it.

    Parameters:
    • module_name: the name of the module to import
    • attr_name: the name of the attribute in the module to be returned
    """

    def callback():
        return getattr(importlib.import_module(module_name), attr_name)

    return lazy(callback)


def is_lazy(obj: Any) -> bool:
    """Return True if an object is a callable is tagged as a lazy callback function."""
    return callable(obj) and getattr(obj, "_fondat__lazy", None)



"""Module for binary content streaming."""

from asyncio import LimitOverrunError
from collections.abc import AsyncIterator
from fondat.validation import MinLen, MinValue, validate_arguments
from typing import Annotated


class Stream(AsyncIterator[bytes | bytearray]):
    """
    Base class to provide binary content through an asynchronous stream. The stream provides
    binary data through asynchronously iterable chunks of bytes or bytearray.

    During iteration, the stream determines the size of each chunk.

    Attributes:
    • content_type: the media (MIME) type of the stream
    • content_length: the length of the content, or None if unknown

    Much like a file, a stream is returned in an "open" state. The consumer must explicitly
    close it, either via by calling its `close` method, or using `async with`.
    """

    def __init__(self, content_type: str, content_length: int | None = None):
        self.content_type = content_type
        self.content_length = content_length

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc):
        await self.close()

    def __aiter__(self):
        return self

    async def __anext__(self) -> bytes | bytearray:
        raise NotImplementedError

    async def close(self) -> None:
        """
        Close the stream. Further attempts to iterate the stream will raise StopAsyncIteration.
        This method is idempotent; it is not an error to close a stream more than once.
        """
        raise NotImplementedError


class BytesStream(Stream):
    """
    Represents a bytes or bytearray object as an asynchronous byte stream. All content is
    returned in a single iteration.

    Parameters:
    • content: the data to be streamed
    • content_type: the MIME type of the data to be streamed
    """

    def __init__(
        self,
        content: bytes | bytearray,
        content_type: str = "application/octet-stream",
    ):
        super().__init__(content_type=content_type, content_length=len(content))
        self._content = content

    async def __anext__(self) -> bytes:
        if self._content is None:
            raise StopAsyncIteration
        result = self._content
        self._content = None
        return result

    async def close(self) -> None:
        self._content = None


class Reader:
    """
    Buffered stream reader.

    If the reader is used as an asynchronous context manager (`async with`), then upon
    exiting the context the stream will be closed.

    Parameters:
    • stream: stream to be read
    • limit: buffer size limit

    If buffer size limit is exceeded during read operations, LimitOverrunError is raised.
    """

    def __init__(self, stream: Stream, limit: int | None = None):
        self.stream = stream
        self.limit = limit
        self._buffer = bytearray()
        self._eof = False

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc):
        await self.close()

    async def _read(self) -> None:
        try:
            self._buffer += await anext(self.stream)
            if self.limit and len(self._buffer) > self.limit:
                raise LimitOverrunError
        except StopAsyncIteration:
            self._eof = True

    @validate_arguments
    async def read(self, size: Annotated[int, MinValue(1)] | None = None) -> bytes:
        """
        Read bytes from the stream.

        Parameter:
        • size: number of bytes to read  [to end of stream]

        This method blocks until all requested bytes are read or the end of the stream is
        encountered.

        The end of stream is signified by zero bytes returned.
        """
        while not self._eof and (size is None or len(self._buffer) < size):
            await self._read()
        result = bytes(self._buffer[:size] if size else self._buffer)
        self._buffer = self._buffer[size:] if size else bytearray(b"")
        return result

    @validate_arguments
    async def read_until(self, separator: Annotated[bytes, MinLen(1)]) -> bytes:
        """
        Read bytes from the stream, up to and including the specified separator.

        Parameter:
        • separator: byte sequence to read up to

        This method blocks until all requested bytes are read or the end of the stream is
        encountered. If it reaches the end of the stream, all bytes read are returned.

        The end of stream is signified by zero bytes returned.
        """
        while not self._eof and separator not in self._buffer:
            await self._read()
        index = self._buffer.find(separator)
        size = index + len(separator) if index != -1 else None
        result = bytes(self._buffer[:size] if size else self._buffer)
        self._buffer = self._buffer[size:] if size else bytearray(b"")
        return result

    async def close(self) -> None:
        """Close the stream."""
        await self.stream.close()


async def read_stream(stream: Stream, limit: int | None = None) -> bytes:
    """Deprecated. Use Reader class."""
    return await Reader(stream, limit).read()



"""Fondat cache module."""

import hashlib
import json

from typing import Any, Protocol, runtime_checkable


JSON = Any


@runtime_checkable
class CacheResource(Protocol):
    """Prototype cache resource."""

    def __getitem__(self, key: bytes) -> "EntryResource":  # subordinate
        ...


@runtime_checkable
class EntryResource(Protocol):
    """Prototype cache entry resource."""

    async def get(self) -> JSON:
        ...

    async def put(self, value: JSON) -> None:
        ...

    async def delete(self) -> None:
        ...


def hash_json(value: JSON) -> bytes:
    """
    Return a deterministic, unique hash value for a given JSON object model value.
    This can be useful for generating a cache entry key.
    """
    return hashlib.sha256(json.dumps(value, sort_keys=True).encode()).digest()
